# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-postgis tests
"""

from cubicweb.devtools import testlib
from cubicweb.devtools import (startpgcluster, stoppgcluster,
                               PostgresApptestConfiguration)

def setUpModule():
    startpgcluster(__file__)

def tearDownModule():
    stoppgcluster(__file__)


class PostgisTC(testlib.CubicWebTC):
    configcls = PostgresApptestConfiguration

    def test_geometry_srid(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('City', name=u'Paris', center_4326=(u'POINT(48.4 2.34)', 4326))
            cnx.commit()
            self.assertEqual(cnx.execute('Any ST_SRID(C) WHERE X center_4326 C')[0][0], 4326)
            # wrong srid
            with self.assertRaises(Exception) as ctx:
                cnx.create_entity('City', name=u'Nantes', center_4326=(u'POINT(47.25 1.5)', -1))
            self.assertEqual(ctx.exception.__class__.__name__, 'DataError')

    def test_geometry_nosrid(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('City', name=u'Paris', center_nosrid=(u'POINT(48.4 2.34)', 4326))
            cnx.commit()
            self.assertEqual(cnx.execute('Any ST_SRID(C) WHERE X center_nosrid C')[0][0], 4326)
            cnx.create_entity('City', name=u'Nantes', center_nosrid=(u'POINT(47.25 1.5)', -1))
            cnx.commit()


if __name__ == '__main__':
    import unittest
    unittest.main()




# USE
# cubicweb-ctl db-create <instance>
# WARNING !!! DO NOT initialize the database
# Add postgis, e.g.:
# psql <database> -f /usr/share/postgresql/9.1/contrib/postgis-1.5/postgis.sql
# psql <database> -f /usr/share/postgresql/9.1/contrib/postgis-1.5/spatial_ref_sys.sql
# cubicweb-ctl db-init <instance>

from logilab.database import FunctionDescr
from logilab.database import get_db_helper
from logilab.database.sqlgen import SQLExpression

from yams import register_base_type

from rql.utils import register_function

from cubes.postgis import _GEOM_PARAMETERS, _GEOG_PARAMETERS

# register new base type
register_base_type('Geometry', _GEOM_PARAMETERS)
register_base_type('Geography', _GEOG_PARAMETERS)


def pg_geometry_sqltype(rdef):
    """Return a PostgreSQL column type corresponding to rdef's geom_type, srid
    and coord_dimension.
    """
    target_geom_type = rdef.geom_type
    if rdef.coord_dimension >= 3:  # XXX: handle 2D+M
        target_geom_type += 'Z'
    if rdef.coord_dimension == 4:
        target_geom_type += 'M'
    assert target_geom_type
    assert rdef.srid
    return 'geometry(%s, %s)' % (target_geom_type, rdef.srid)

def pg_geography_sqltype(rdef):
    """Return a PostgreSQL column type corresponding to rdef's geom_type and srid
    """
    srid = rdef.srid or 4326
    return 'geography(%s, %s)' % (rdef.geom_type, srid)

# Add the datatype to the helper mapping
pghelper = get_db_helper('postgres')
pghelper.TYPE_MAPPING['Geometry'] = pg_geometry_sqltype
pghelper.TYPE_MAPPING['Geography'] = pg_geography_sqltype

# Add a converter for Geometry
def convert_geom(x):
    if isinstance(x, (tuple, list)):
        # We give the (Geometry, SRID)
        return SQLExpression('ST_GeomFromText(%(geo)s, %(srid)s)', geo=x[0], srid=x[1])
    else:
        # We just give the Geometry
        return SQLExpression('ST_GeomFromText(%(geo)s, %(srid)s)', geo=x, srid=-1)

def convert_geog(x):
    # takes only a Geometry type, assumes GPS srid
    return SQLExpression('ST_GeogFromText(%(geo)s)', geo=x)

# Add the converter function to the known SQL_CONVERTERS
pghelper.TYPE_CONVERTERS['Geometry'] = convert_geom
pghelper.TYPE_CONVERTERS['Geography'] = convert_geog


class ST_EQUALS(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'


class ST_INTERSECTS(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'

class ST_INTERSECTION(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_OVERLAPS(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'

class ST_CROSSES(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'

class ST_TOUCHES(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'


class ST_GEOMETRYN(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_WITHIN(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'


class ST_CONTAINS(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'


class ST_DWITHIN(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_LENGTH(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_DISTANCE(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_DISTANCE_SPHERE(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_AREA(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_NUMINTERIORRINGS(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Int'


class ST_SIMPLIFY(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_TRANSFORM(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_ASGEOJSON(FunctionDescr):
    minargs = 1
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'String'


class ST_ASTEXT(FunctionDescr):
    minargs = 1
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'String'


class ST_GEOMFROMTEXT(FunctionDescr):
    minargs = 1
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'

class ST_GEOGFROMTEXT(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geography'

class ST_COVERS(FunctionDescr):
    minargs = 1
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Bool'

class GEOMETRY(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geometry'

class ST_UNION(FunctionDescr):
    aggregat = True
    minargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geometry'

class ST_COLLECT(FunctionDescr):
    aggregat = True
    minargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geometry'

class ST_CONVEXHULL(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geometry'

class ST_CONCAVEHULL(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class DISPLAY(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'String'

    def as_sql_postgres(self, args):
        return 'ST_ASGeoJSON(ST_Transform(%s, %s))' % (args[0], args[1])


class ST_MAKEPOINT(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_SRID(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_SETSRID(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_X(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_Y(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_Z(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Float'


class ST_M(FunctionDescr):
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Float'


register_function(ST_EQUALS)
register_function(ST_INTERSECTS)
register_function(ST_INTERSECTION)
register_function(ST_OVERLAPS)
register_function(ST_CROSSES)
register_function(ST_TOUCHES)
register_function(ST_GEOMETRYN)
register_function(ST_WITHIN)
register_function(ST_CONTAINS)
register_function(ST_DWITHIN)
register_function(ST_LENGTH)
register_function(ST_DISTANCE)
register_function(ST_AREA)
register_function(ST_NUMINTERIORRINGS)
register_function(ST_SIMPLIFY)
register_function(ST_TRANSFORM)
register_function(DISPLAY)
register_function(ST_ASGEOJSON)
register_function(ST_ASTEXT)
register_function(ST_GEOMFROMTEXT)
register_function(ST_GEOGFROMTEXT)
register_function(ST_COVERS)
register_function(GEOMETRY)
register_function(ST_UNION)
register_function(ST_COLLECT)
register_function(ST_CONVEXHULL)
register_function(ST_CONCAVEHULL)
register_function(ST_MAKEPOINT)
register_function(ST_SETSRID)
register_function(ST_SRID)
register_function(ST_DISTANCE_SPHERE)
register_function(ST_X)
register_function(ST_Y)
register_function(ST_Z)
register_function(ST_M)



"""cubicweb-postgis application package

Wrap Postgis into CubicWeb.
"""

_GEOM_PARAMETERS = ('srid', 'geom_type', 'coord_dimension')  # XXX: coord_dimension -> z and m ?
_GEOG_PARAMETERS = ('srid', 'geom_type', 'coord_dimension')



# copyright 2012-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-postgis precreate script, executed at instance creation time or when
the cube is added to an existing instance.

You can add custom actions here before first entities are inserted into database.
"""



# -*- coding: utf-8 -*-
# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-postgis views/forms/actions/components for web ui"""
from cwtags.tag import div, label, input

from cubicweb import uilib
from cubicweb.utils import UStringIO
from cubicweb.web import formfields
from cubicweb.web import formwidgets
from cubicweb.view import EntityView
from cubicweb.predicates import match_kwargs

try:
    from shapely import wkb
except ImportError:
    wkb = None


# StringField is probably not the worst default choice
formfields.FIELDS['Geometry'] = formfields.StringField
uilib.PRINTERS['Geometry'] = uilib.print_string
formfields.FIELDS['Geography'] = formfields.StringField
uilib.PRINTERS['Geography'] = uilib.print_string


class RefpointAttributeView(EntityView):
    __regid__ = 'refpointattr'
    __select__ = EntityView.__select__ & match_kwargs('rtype')

    def entity_call(self, entity, rtype, **kwargs):
        value = getattr(entity, rtype)
        if value:
            value = wkb.loads(value.decode('hex'))
            self.w(u'lat: %s, long: %s' % (value.x, value.y))


class RefpointWidget(formwidgets.TextInput):

    def _render(self, form, field, renderer):
        ustring = UStringIO()
        w = ustring.write
        value = self.values(form, field)[0]
        longitude = 0
        latitude = 0
        if value:
            value = wkb.loads(value.decode('hex'))
            longitude = value.x
            latitude = value.y
        with div(w, class_='form-group'):
            w(label(u'longitude'))
            w(input(type='text', value=longitude,
                    name=field.input_name(form, 'longitude')))
        with div(w, class_='form-group'):
            w(label(u'latitude'))
            w(input(type='text', value=latitude,
                    name=field.input_name(form, 'latitude')))
        return ustring.getvalue()

    def process_field_data(self, form, field):
        """Return process posted value(s) for widget and return something
        understandable by the associated `field`. That value may be correctly
        typed or a string that the field may parse.
        """
        posted = form._cw.form
        longitude = posted.get(field.input_name(form, 'longitude')).strip() or None
        latitude =  posted.get(field.input_name(form, 'latitude')).strip() or None
        return u'POINT(%s %s)' % (longitude, latitude)



# pylint: disable=W0622
"""cubicweb-postgis application packaging information"""

modname = 'postgis'
distname = 'cubicweb-postgis'

numversion = (0, 4, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'Test for postgis'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ =  {'cubicweb': '>= 3.22.0',
                'cwtags': '>= 1.1.0',
                }
__recommends__ = {}


from os import listdir as _listdir
from os.path import join, isdir
from glob import glob

THIS_CUBE_DIR = join('share', 'cubicweb', 'cubes', modname)

def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc')
            and not fname.endswith('~')
            and not isdir(join(dirpath, fname))]

data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
    ]
# check for possible extended cube layout
for dname in ('entities', 'views', 'sobjects', 'hooks', 'schema', 'data', 'wdoc', 'i18n', 'migration'):
    if isdir(dname):
        data_files.append([join(THIS_CUBE_DIR, dname), listdir(dname)])
# Note: here, you'll need to add subdirectories if you want
# them to be included in the debian package

